'''Plays MIDILike Objects'''
import threading
import time
import os
import pyinotify

from apres import MIDI, MIDIController, MIDIEvent, NoteOn, NoteOff
from rory.midiinterface import MIDIInterface

class Player:
    '''Plays MIDILike Objects'''

    def kill(self):
        ''''Shutdown the player'''
        self.is_active = False
        self.midi_controller.close()

    def next_state(self):
        '''Change the song position to the next state with notes.'''
        new_position = self.song_position + 1
        while new_position <= self.loop[1] \
        and not self.midi_interface.get_state(new_position, self.ignored_channels):
            new_position += 1

        new_position = min(self.loop[1] + 1, new_position)

        if new_position > self.loop[1]:
            new_position = self.loop[0]
            self.song_position = new_position
        else:
            self.song_position = new_position


    def prev_state(self):
        '''Change the song position to the last state with notes.'''
        self.song_position -= 1
        while self.song_position > self.loop[0] \
        and not self.midi_interface.get_state(self.song_position, self.ignored_channels):
            self.song_position -= 1

        self.set_state(max(0, self.song_position))

    def set_state(self, song_position):
        '''
            Set the song position as the value in the register,
            then move to the next state with notes.
        '''
        self.song_position = max(0, song_position)

        while self.song_position < self.loop[1] \
        and not self.midi_interface.get_state(self.song_position, self.ignored_channels):
            self.song_position += 1

        self.song_position = min(self.loop[1], self.song_position)

        if self.song_position == self.loop[1]:
            self.song_position = self.loop[0]

    def set_measure(self, measure):
        position = self.midi_interface.get_first_position_in_measure(measure)
        self.set_state(position)

    def reinit_midi_interface(self, **kwargs):
        self.active_midi = MIDI(self.active_midi.path)
        self.midi_interface = MIDIInterface(self.active_midi, **kwargs)
        self.clear_loop()
        self.song_position = -1
        self.next_state()

    def get_transpose(self):
        return self.midi_interface.transpose

    def __init__(self, **kwargs):
        self.active_midi = MIDI(kwargs['path'])

        self.is_active = True
        self.register = 0
        self.flag_negative_register = False
        self.loop = [0, 0]
        self.note_range = (21, 109)

        self.ignored_channels = set()

        self.midi_interface = MIDIInterface(self.active_midi, **kwargs)

        self.clear_loop()

        self.need_to_release = set()

        self.song_position = -1

        self.next_state()

        self.flag_range_input = False
        self._new_range = None

        if 'controller_path' in kwargs:
            self.midi_controller = RoryController(self, kwargs['controller_path'])
        else:
            self.midi_controller = RoryController(self)

    def get_register(self):
        if self.flag_negative_register:
            self.register *= -1

        output = self.register
        self.register = 0
        self.flag_negative_register = False
        return output

    def set_note_range(self, lower, upper):
        self.note_range = (lower, upper)
        self.flag_range_input = False
        self._new_range = None

    def get_pressed_notes(self):
        ''' Get the notes that the midi device has held down '''
        notes = self.midi_controller.pressed.copy()
        if self.flag_range_input:
            if len(notes):
                lower = min(notes)
                upper = max(notes)
                if self._new_range is None:
                    self._new_range = (lower, upper)
                else:
                    self._new_range = (
                        min(self._new_range[0], lower),
                        max(self._new_range[1], upper)
                    )

                if self._new_range[0] != self._new_range[1]:
                    self.set_note_range(*self._new_range)
        return notes

    def flag_new_range(self):
        self.flag_range_input = True

    def do_state_check(self):
        ''' Check if the midi device is pressing the coresponding notes '''
        song_state = self.midi_interface.get_state(self.song_position, self.ignored_channels)
        pressed = self.get_pressed_notes()
        if song_state.intersection(pressed) == song_state \
        and not self.need_to_release.intersection(song_state):
            self.need_to_release = self.need_to_release.union(pressed)
            self.next_state()

    def toggle_ignore_channel(self, channel):
        '''
            Add or remove a channel to be ignored when considering
            if the song position needs to be incremented or decremented
        '''
        if channel < 16:
            if channel in self.ignored_channels:
                self.ignored_channels.remove(channel)
            else:
                self.ignored_channels.add(channel)

        self.set_state(self.song_position)

    def unignore_channel(self, channel):
        try:
            self.ignored_channels.remove(channel)
        except KeyError:
            pass

    def ignore_channel(self, channel):
        self.ignored_channels.add(channel)

    def set_loop_start_to_position(self):
        '''Set the beginning of the play loop to the current song position'''
        self.set_loop_start(self.song_position)

    def set_loop_end_to_position(self):
        '''Set the end of the play loop to the current song position'''
        self.set_loop_end(self.song_position)

    def set_loop_start(self, position):
        '''set current positions as loop start'''
        self.loop[0] = min(max(0, position), len(self.midi_interface.state_map) - 1)

    def set_loop_end(self, position):
        '''set current positions as loop end'''
        self.loop[1] = min(max(0, position), len(self.midi_interface.state_map) - 1)

    def clear_loop(self):
        '''Stop Looping'''
        self.loop = [0, len(self.midi_interface.state_map) - 1]

    def set_register_digit(self, digit):
        '''Insert digit to register'''
        if digit == ord('-'):
            self.register = 0
            self.flag_negative_register = True
        else:
            assert (digit < 10), "Digit can't be more than 9. Called from somewhere it shouldn't be"
            self.register *= 10
            self.register += digit

    def clear_register(self):
        '''Set the register to 0'''
        self.register = 0
        self.flag_negative_register = False

    def jump_to_register_position(self):
        '''Set the song position to the value of the input register'''
        self.set_state(self.get_register())


class RoryController(MIDIController):
    def __init__(self, player, path = ""):
        self.player = player

        if not path:
            for filename in os.listdir("/dev/"):
                if "midi" in filename:
                    path = "/dev/%s" % filename
                    break

        super().__init__(path)

        self.pressed = set()

        self.watch_manager = pyinotify.WatchManager()
        notifier = pyinotify.ThreadedNotifier(self.watch_manager, TaskHandler(self))
        notifier.daemon = True
        notifier.start()
        self.watch_manager.add_watch(
            "/dev/",
            pyinotify.IN_CREATE | pyinotify.IN_DELETE
        )

        self.state_check_ticket = 0
        self.processing_ticket = 0

    def do_state_check(self):
        ''' Ticketed wrapper for player's do_state_check '''
        my_ticket = self.state_check_ticket
        self.state_check_ticket += 1

        while my_ticket != self.processing_ticket:
            time.sleep(.05)

        # If there are newer tickets queued, skip this state_check
        if my_ticket == self.state_check_ticket - 1:
            self.player.do_state_check()
        else:
            pass

        self.processing_ticket += 1

    def hook_NoteOn(self, event):
        if event.velocity == 0:
            self.release_note(event.note)
        else:
            self.press_note(event.note)

    def hook_NoteOff(self, event):
        self.release_note(event.note)

    def press_note(self, note):
        '''Press a Midi Note'''
        self.pressed.add(note)
        self.do_state_check()

    def release_note(self, note):
        '''Release a Midi Note'''
        try:
            self.pressed.remove(note)
        except KeyError:
            pass
        try:
            self.player.need_to_release.remove(note)
        except KeyError:
            pass
        self.do_state_check()

    def connect(self, path):
        self.player.need_to_release = set()
        try:
            super().connect(path)
            # Automatically start listening for input on connect
            input_thread = threading.Thread(
                target=self.listen
            )
            input_thread.start()
        except FileNotFoundError:
            pass

class TaskHandler(pyinotify.ProcessEvent):
    '''Event hooks to connect/disconnect from newly made midi device'''
    def __init__(self, controller):
        self.controller = controller
        super().__init__()

    def process_IN_CREATE(self, event):
        '''Hook to connect when midi device is plugged in'''
        if event.name[0:4] == 'midi':
            time.sleep(.5)
            self.controller.connect(event.pathname)

    def process_IN_DELETE(self, event):
        '''Hook to disconnect when midi device is unplugged'''
        if self.controller.midipath == event.pathname:
            self.controller.disconnect()




'''Plays MIDILike Objects'''
import math
from apres import NoteOn, NoteOff, SetTempo, TimeSignature, MIDIEvent
from rory.structures import Grouping

class MIDIInterface:
    '''Layer between Player and the MIDI input file'''
    notelist = 'CCDDEFFGGAAB'

    def __handle_kwargs(self, kwargs):
        if 'transpose' in kwargs:
            self.transpose = kwargs['transpose']

    def __calculate_beat_chunks(self):
        beats = []

        final_tick = 0 # will be the final tick with a note on event
        running_beat_count = (0, 0) # beat_count, last_tick_totalled

        current_numerator = 4
        beat_in_measure = 0
        current_measure = 0
        beat_size = self.midi.ppqn
        active_notes = {}
        min_note = 128
        max_note = 0
        for tick, event in self.midi.get_all_events():
            final_tick = tick
            tick_diff = tick - running_beat_count[1]

            current_beat = int(running_beat_count[0] + (tick_diff // beat_size))
            while len(beats) <= current_beat:
                beats.append([[], beat_size, current_measure, current_numerator, beat_in_measure])
                beat_in_measure += 1
                if beat_in_measure == current_numerator:
                    beat_in_measure = 0
                    current_measure += 1

            if isinstance(event, NoteOn) and event.channel != 9 and event.velocity > 0:
                active_notes[(event.channel, event.note)] = (current_beat, len(beats[current_beat][0]))
                beats[current_beat][0].append((tick_diff % beat_size, event, tick, 0))
                min_note = min(min_note, event.note)
                max_note = max(max_note, event.note)

            elif (isinstance(event, NoteOn) and event.channel != 9 and event.velocity == 0) or (isinstance(event, NoteOff) and event.channel != 9):
                try:
                    beat, index = active_notes[(event.channel, event.note)]
                    _a, _b, original_tick, duration = beats[beat][0][index]
                    beats[beat][0][index] = (_a, _b, original_tick, tick - original_tick)
                except KeyError:
                    pass

            elif isinstance(event, TimeSignature):
                running_beat_count = (current_beat, tick)
                current_numerator = event.numerator
                beat_size = self.midi.ppqn // ((2 ** event.denominator) / 4)
                beats[current_beat][1] = beat_size
                beats[current_beat][3] = current_numerator

        if final_tick != running_beat_count[1]:
            tick_diff = final_tick - running_beat_count[1]
            #for i in range(math.ceil(tick_diff / (beat_size * current_numerator))):
            #    measure_map.append((i * (beat_size * current_numerator)) + running_beat_count[1])

        self.transpose = min(
            max(
                self.transpose,
                min_note * -1
            ),
            128 - max_note
        )

        return beats

    def __beats_to_grouping(self, beats):
        grouping = Grouping()
        measures = []
        for (_events, _beat_size, measure, _numerator, _bim) in beats:
            while len(measures) <= measure:
                measures.append(measure)

        grouping.set_size(len(measures))
        for _events, _beat_size, m_index, numerator, _bim in beats:
            measure = grouping.get_grouping(m_index)
            measure.set_size(numerator)

        for events, beat_size, m_index, _numerator, bim in beats:
            measure = grouping.get_grouping(m_index)
            for (pos, event, _real, _duration) in events:
                beat = measure.get_grouping(bim)
                beat.set_size(beat_size)

            for (pos, event, real, _duration) in events:
                beat = measure.get_grouping(bim)
                tick = beat.get_grouping(int(pos))
                tick.add_event((event, real))

        return grouping

    def __init__(self, midi, **kwargs):
        self.midi = midi

        # For quick access to which keys are pressed
        self.state_map = []
        self.active_notes_map = []
        self.beat_map = {}
        self.inv_beat_map = {}
        self.rhythm_map = {0: (0, 1)}
        self.measure_map = [] # { state_position: measure_number }
        self.timing_map = {0: 0}
        self.transpose = 0

        self.__handle_kwargs(kwargs)

        beats = self.__calculate_beat_chunks()
        grouping = self.__beats_to_grouping(beats)

        beat_count = 0
        for measure in grouping.iter():
            self.measure_map.append(len(self.state_map))
            for beat_index, beat in enumerate(measure.iter()):
                beat.reduce()
                self.beat_map[len(self.state_map)] = beat_count
                self.inv_beat_map[beat_count] = len(self.state_map)

                l, flat_min = beat.get_flat_min()

                if len(flat_min) / l >= .50:
                    frames_to_add = l
                else:
                    frames_to_add = int(len(flat_min) * 1.5)

                frames_to_add = max(1, frames_to_add)

                initial_i = len(self.state_map)
                for i in range(frames_to_add):
                    self.state_map.append(set())
                    self.active_notes_map.append({})

                for index, eventlist in flat_min:
                    i = int(initial_i + ((index / l) * frames_to_add))
                    for event, real_tick in eventlist:
                        new_note = event.note + self.transpose
                        event.set_note(new_note)
                        self.state_map[i].add(event.note)
                        self.active_notes_map[i][event.note] = event
                        self.timing_map[i] = real_tick
                beat_count += 1



    def get_real_tick(self, song_position):
        ''' Get the tick from before the midi is processed for playing '''
        first_post = song_position
        last_post = song_position
        keys = self.timing_map.keys()
        max_tick = max(keys)
        min_tick = min(keys)

        divs = 0
        while first_post not in keys and first_post > min_tick:
            first_post -= 1
            divs += 1

        while last_post not in keys and last_post <= max_tick:
            last_post += 1
            divs += 1

        if last_post > max_tick:
            diff = len(self.midi) - self.timing_map[first_post]
        else:
            diff = self.timing_map[last_post] - self.timing_map[first_post]

        if divs:
            diff //= divs

        return diff + self.timing_map[first_post]

    def get_tick_wait(self, song_position, new_position):
        ''' Calculate how long, in midi ticks, between to song positions '''
        first_post = song_position
        last_post = new_position
        keys = self.timing_map.keys()
        max_tick = max(keys)
        min_tick = min(keys)

        divs = 0
        while first_post not in keys and first_post > min_tick:
            first_post -= 1
            divs += 1

        while last_post not in keys and last_post <= max_tick:
            last_post += 1
            divs += 1

        if last_post > max_tick:
            diff = len(self.midi) - self.timing_map[first_post]
        else:
            diff = self.timing_map[last_post] - self.timing_map[first_post]

        if divs:
            diff //= divs

        return diff


    def get_state(self, position, ignored_channels = None):
        '''Get a list of the notes currently 'On' at specified position'''
        if not ignored_channels:
            state = self.state_map[position].copy()
        else:
            state = set()
            for note, event in self.active_notes_map[position].items():
                if event.channel not in ignored_channels:
                    state.add(note)

        return state

    def get_active_channels(self, position):
        ''' Get set of channels present at a given position '''
        active = set()
        for _note, event in self.active_notes_map[position].items():
            active.add(event.channel)

        return active

    def get_chord_name(self, position, channel):
        ''' Attempt to detect the name of the chord being played at a given position '''
        chord_names = {
            (0, 3, 7): "m",
            (0, 4, 9): "m/1",
            (0, 5, 8): "m/2",

            (0, 4, 7): "", # Major
            (0, 3, 8): "/1",
            (0, 5, 9): "/2",

            (0, 3, 6): "dim",
            (0, 3, 9): "dim/1",
            (0, 6, 9): "dim/2",

            (0, 3, 6, 9): "dim7",

            (0, 4, 8): "aug", # Symmetrical
            (0, 4, 8, 10): "aug7",

            (0, 2, 7): "sus2",
            (0, 5, 10): "sus2/1", # sus4/2

            (0, 2, 7, 10): "7sus2",
            (0, 5, 7): "sus4", # sus2/2

            (0, 6, 7, 10): "7sus4",
            #(0, 12): "open",
            (0, 4): "3",
            (0, 3): "m3",
            (0, 9): "m3/1",

            (0, 7): "5",
            (0, 5): "5/1",
            (0, 8): "aug5", #3/1
            (0, 6): "dim5", #3/1

            (0, 4, 7, 9): "6",
            (0, 3, 7, 9): "m6",
            (0, 4, 7, 10): "7",
            (0, 3, 7, 10): "m7",
            (0, 3, 6, 10): "m7b5",
            (0, 4, 7, 11): "maj7",
            (0, 3, 7, 11): "mM7",
            (0, 4, 6, 10): "7-5",
            (0, 4, 8, 10): "7+5",
            (0, 4, 7, 10, 14): "9",
            (0, 3, 7, 10, 14): "m9",
            (0, 4, 7, 11, 14): "maj9",
            (0, 4, 7, 10, 14, 17): "11",
            (0, 3, 7, 10, 14, 17): "m11",
            (0, 2, 4, 7): "add2",
            (0, 4, 5, 7): "add4",
            (0, 4, 7, 14): "add9",
            (0, 4, 7, 17): "add11",
        }


        pressed = []
        for note, event in self.active_notes_map[position].items():
            if event.channel == channel:
                pressed.append(note)

        tonic = min(pressed)
        for i, pressed_note in enumerate(pressed):
            pressed[i] = (pressed_note - tonic) % 12
        pressed = list(set(pressed))

        pressed.sort()

        pressed = tuple(pressed)
        if pressed in chord_names:
            name = chord_names[pressed]

            if name[-2:] == "/1":
                slash = self.get_note_name(tonic)
                tonic += pressed[-1]
                name = "%s%s%s" % (self.get_note_name(tonic), name[0:-1], slash)
            elif name[-2:] == "/2":
                slash = self.get_note_name(tonic)
                tonic += pressed[-2]
                name = "%s%s%s" % (self.get_note_name(tonic), name[0:-1], slash)
            else:
                name = self.get_note_name(tonic) + name
        else:
            name = ""

        return name

    def __len__(self):
        return len(self.state_map)

    def get_first_position_in_measure(self, measure):
        measure = min(measure, max(len(self.measure_map) - 1, 0))
        return self.measure_map[measure]


    def get_measure(self, test_position):
        output = 0

        for (i, relative_position) in enumerate(self.measure_map):
            if test_position < relative_position:
                break
            output = i

        return output

    def get_beat(self, test_position):
        return self.inv_beat_map[test_position]

    @staticmethod
    def get_note_name(midi_note):
        ''' Get note's letter name '''
        name = MIDIInterface.notelist[midi_note % len(MIDIInterface.notelist)]
        if midi_note % len(MIDIInterface.notelist) in (1,3,6,8,10):
            name += "#"

        return name
